import time as ttime
import subprocess
import pytest
import os
import yaml

from bluesky_queueserver.manager.profile_ops import gen_list_of_plans_and_devices
from bluesky_queueserver.manager.comms import generate_new_zmq_key_pair

from .common import (
    patch_first_startup_file,
    patch_first_startup_file_undo,
    wait_for_condition,
    condition_manager_idle,
    condition_manager_paused,
    condition_environment_created,
    condition_environment_closed,
    condition_queue_processing_finished,
    get_reduced_state_info,
    get_queue_state,
    get_queue,
    append_code_to_last_startup_file,
    set_qserver_zmq_public_key,
    zmq_single_request,
)

from .common import re_manager, re_manager_pc_copy, re_manager_cmd  # noqa: F401

from ..qserver_cli import QServerExitCodes

# Exit codes for CLI tool
SUCCESS = QServerExitCodes.SUCCESS.value
PARAM_ERROR = QServerExitCodes.PARAMETER_ERROR.value
REQ_FAILED = QServerExitCodes.REQUEST_FAILED.value
COM_ERROR = QServerExitCodes.COMMUNICATION_ERROR.value
EXCEPTION_OCCURRED = QServerExitCodes.EXCEPTION_OCCURRED.value

timeout_env_open = 10


def test_qserver_cli_and_manager(re_manager):  # noqa: F811
    """
    Long test runs a series of CLI commands.
    """
    assert wait_for_condition(
        time=3, condition=condition_manager_idle
    ), "Timeout while waiting for manager to initialize."

    # Clear queue
    assert subprocess.call(["qserver", "queue", "clear"]) == SUCCESS

    # Request the list of allowed plans and devices (we don't check what is returned)
    assert subprocess.call(["qserver", "allowed", "plans"], stdout=subprocess.DEVNULL) == SUCCESS
    assert subprocess.call(["qserver", "allowed", "devices"], stdout=subprocess.DEVNULL) == SUCCESS

    # Request the list of existing plans and devices (we don't check what is returned)
    assert subprocess.call(["qserver", "existing", "plans"], stdout=subprocess.DEVNULL) == SUCCESS
    assert subprocess.call(["qserver", "existing", "devices"], stdout=subprocess.DEVNULL) == SUCCESS

    # Add a number of plans
    plan_1 = "{'name':'count', 'args':[['det1', 'det2']]}"
    plan_2 = "{'name':'scan', 'args':[['det1', 'det2'], 'motor', -1, 1, 10]}"
    plan_3 = "{'name':'count', 'args':[['det1', 'det2']], 'kwargs':{'num':10, 'delay':1}}"
    assert subprocess.call(["qserver", "queue", "add", "plan", plan_1]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "add", "plan", plan_2]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "add", "plan", plan_3]) == SUCCESS

    n_plans, is_plan_running, _ = get_reduced_state_info()
    assert n_plans == 3, "Incorrect number of plans in the queue"
    assert not is_plan_running, "Plan is executed while it shouldn't"

    assert subprocess.call(["qserver", "queue", "get"]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "item", "remove"]) == SUCCESS

    n_plans, is_plan_running, _ = get_reduced_state_info()
    assert n_plans == 2, "Incorrect number of plans in the queue"

    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS

    assert wait_for_condition(time=60, condition=condition_queue_processing_finished)

    # Smoke test for 'history_get' and 'history_clear'
    assert subprocess.call(["qserver", "history", "get"]) == SUCCESS
    assert subprocess.call(["qserver", "history", "clear"]) == SUCCESS

    # Queue is expected to be empty (processed). Load one more plan.
    assert subprocess.call(["qserver", "queue", "add", "plan", plan_3]) == SUCCESS

    n_plans, is_plan_running, _ = get_reduced_state_info()
    assert n_plans == 1, "Incorrect number of plans in the queue"

    assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
    ttime.sleep(1)
    assert subprocess.call(["qserver", "re", "pause", "immediate"]) == SUCCESS
    assert wait_for_condition(
        time=60, condition=condition_manager_paused
    ), "Timeout while waiting for manager to pause"

    assert subprocess.call(["qserver", "re", "resume"]) == SUCCESS
    ttime.sleep(1)
    assert subprocess.call(["qserver", "re", "pause", "deferred"]) == SUCCESS
    assert wait_for_condition(
        time=60, condition=condition_manager_paused
    ), "Timeout while waiting for manager to pause"

    assert subprocess.call(["qserver", "re", "resume"]) == SUCCESS

    assert wait_for_condition(
        time=60, condition=condition_queue_processing_finished
    ), "Timeout while waiting for process to finish"

    assert subprocess.call(["qserver", "queue", "add", "plan", plan_1]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "add", "plan", plan_1]) == SUCCESS

    n_plans, is_plan_running, _ = get_reduced_state_info()
    assert n_plans == 2, "Incorrect number of plans in the queue"

    assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS

    assert wait_for_condition(
        time=60, condition=condition_queue_processing_finished
    ), "Timeout while waiting for process to finish"

    # Test 'killing' the manager during running plan. Load long plan and two short ones.
    #   The tests checks if execution of the queue is continued uninterrupted after
    #   the manager restart
    assert subprocess.call(["qserver", "queue", "add", "plan", plan_3]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "add", "plan", plan_3]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "add", "plan", plan_3]) == SUCCESS
    n_plans, is_plan_running, _ = get_reduced_state_info()
    assert n_plans == 3, "Incorrect number of plans in the queue"

    assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
    ttime.sleep(1)
    assert subprocess.call(["qserver", "manager", "kill", "test"]) != SUCCESS
    ttime.sleep(6)  # Don't request the condition to avoid timeout error TODO: wait for the server
    assert wait_for_condition(
        time=60, condition=condition_queue_processing_finished
    ), "Timeout while waiting for process to finish"

    assert subprocess.call(["qserver", "environment", "close"]) == SUCCESS
    assert wait_for_condition(
        time=5, condition=condition_environment_closed
    ), "Timeout while waiting for environment to be closed"


def test_qserver_environment_close(re_manager):  # noqa: F811
    """
    Test for `environment_close` command
    """
    assert wait_for_condition(
        time=3, condition=condition_manager_idle
    ), "Timeout while waiting for manager to initialize."

    # Clear queue
    assert subprocess.call(["qserver", "queue", "clear"]) == SUCCESS

    plan = "{'name':'count', 'args':[['det1', 'det2']], 'kwargs':{'num':5, 'delay':1}}"
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS

    n_plans, is_plan_running, _ = get_reduced_state_info()
    assert n_plans == 1, "Incorrect number of plans in the queue"
    assert is_plan_running is False

    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
    ttime.sleep(2)
    n_plans, is_plan_running, _ = get_reduced_state_info()
    assert n_plans == 0, "Incorrect number of plans in the queue"
    assert is_plan_running is True

    # Call is expected to fail, because a plan is currently running
    assert subprocess.call(["qserver", "environment", "close"]) != SUCCESS

    assert wait_for_condition(
        time=60, condition=condition_queue_processing_finished
    ), "Timeout while waiting for process to finish"

    n_plans, is_plan_running, n_history = get_reduced_state_info()
    assert n_plans == 0, "Incorrect number of plans in the queue"
    assert is_plan_running is False
    assert n_history == 1

    # Now we can close the environment because plan execution is complete
    assert subprocess.call(["qserver", "environment", "close"]) == SUCCESS
    assert wait_for_condition(
        time=5, condition=condition_environment_closed
    ), "Timeout while waiting for environment to be closed"


def test_qserver_environment_destroy(re_manager):  # noqa: F811
    """
    Test for `environment_destroy` command
    """
    assert wait_for_condition(
        time=3, condition=condition_manager_idle
    ), "Timeout while waiting for manager to initialize."

    # Clear queue
    assert subprocess.call(["qserver", "queue", "clear"]) == SUCCESS

    plan = "{'name':'count', 'args':[['det1', 'det2']], 'kwargs':{'num':5, 'delay':1}}"
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS

    n_plans, is_plan_running, _ = get_reduced_state_info()
    assert n_plans == 1, "Incorrect number of plans in the queue"
    assert is_plan_running is False

    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
    ttime.sleep(2)
    n_plans, is_plan_running, _ = get_reduced_state_info()
    assert n_plans == 0, "Incorrect number of plans in the queue"
    assert is_plan_running is True

    assert subprocess.call(["qserver", "environment", "destroy"]) == SUCCESS
    assert wait_for_condition(
        time=3, condition=condition_manager_idle
    ), "Timeout while waiting for environment to be destroyed."

    n_plans, is_plan_running, _ = get_reduced_state_info()
    assert n_plans == 1, "Incorrect number of plans in the queue"
    assert is_plan_running is False

    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
    ttime.sleep(2)
    n_plans, is_plan_running, _ = get_reduced_state_info()
    assert n_plans == 0, "Incorrect number of plans in the queue"
    assert is_plan_running is True

    assert wait_for_condition(
        time=60, condition=condition_queue_processing_finished
    ), "Timeout while waiting for process to finish"

    n_plans, is_plan_running, n_history = get_reduced_state_info()
    assert n_plans == 0, "Incorrect number of plans in the queue"
    assert is_plan_running is False
    assert n_history == 2

    assert subprocess.call(["qserver", "environment", "close"]) == SUCCESS
    assert wait_for_condition(
        time=5, condition=condition_environment_closed
    ), "Timeout while waiting for environment to be closed"


# fmt: off
@pytest.mark.parametrize("option_pause, option_continue", [
    ("deferred", "resume"),
    ("immediate", "resume"),
    ("deferred", "stop"),
    ("deferred", "abort"),
    ("deferred", "halt")
])
# fmt: on
def test_qserver_re_pause_continue(re_manager, option_pause, option_continue):  # noqa: F811
    """
    Test for `re_pause`, `re_resume`, `re_stop`, `re_abort` and `re_halt` commands
    """
    assert wait_for_condition(
        time=3, condition=condition_manager_idle
    ), "Timeout while waiting for manager to initialize."

    # Out of place calls
    assert subprocess.call(["qserver", "re", option_continue]) == REQ_FAILED
    assert subprocess.call(["qserver", "re", "pause", option_pause]) == REQ_FAILED

    # Clear queue
    assert subprocess.call(["qserver", "queue", "clear"]) == SUCCESS

    plan = "{'name':'count', 'args':[['det1', 'det2']], 'kwargs':{'num': 10, 'delay': 1}}"
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS

    n_plans, is_plan_running, _ = get_reduced_state_info()
    assert n_plans == 2, "Incorrect number of plans in the queue"
    assert is_plan_running is False

    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
    ttime.sleep(2)

    # Out of place calls
    assert subprocess.call(["qserver", "re", option_continue]) == REQ_FAILED

    assert subprocess.call(["qserver", "re", "pause", option_pause]) == SUCCESS
    assert wait_for_condition(
        time=3, condition=condition_manager_paused
    ), "Timeout while waiting for manager to pause"

    status = get_queue_state()
    assert status["manager_state"] == "paused"

    n_plans, is_plan_running, n_history = get_reduced_state_info()
    assert n_plans == 1, "Incorrect number of plans in the queue"
    assert is_plan_running is False
    assert n_history == 0

    # Out of place call
    assert subprocess.call(["qserver", "re", "pause", option_pause]) == REQ_FAILED

    assert subprocess.call(["qserver", "re", option_continue]) == SUCCESS

    if option_continue == "resume":
        n_history_expected = 2
    else:
        assert wait_for_condition(time=3, condition=condition_manager_idle)

        n_plans, is_plan_running, n_history = get_reduced_state_info()
        assert n_plans == 2, "Incorrect number of plans in the queue"
        assert is_plan_running is False
        assert n_history == 1

        assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS

        n_history_expected = 3  # Includes entry related to 1 stopped plan

    ttime.sleep(1)

    n_plans, is_plan_running, n_history = get_reduced_state_info()
    assert n_plans == 1, "Incorrect number of plans in the queue"
    assert is_plan_running is True
    assert n_history == n_history_expected - 2

    assert wait_for_condition(
        time=60, condition=condition_queue_processing_finished
    ), "Timeout while waiting for process to finish"

    n_plans, is_plan_running, n_history = get_reduced_state_info()
    assert n_plans == 0, "Incorrect number of plans in the queue"
    assert is_plan_running is False
    assert n_history == n_history_expected

    assert subprocess.call(["qserver", "environment", "close"]) == SUCCESS
    assert wait_for_condition(
        time=5, condition=condition_environment_closed
    ), "Timeout while waiting for environment to be closed"


# fmt: off
@pytest.mark.parametrize("time_kill", ["before", 2, 8, "paused"])
# fmt: on
def test_qserver_manager_kill(re_manager, time_kill):  # noqa: F811
    """
    Test for `test_manager_kill` command. The command is stopping the event loop of RE Manager,
    causeing RE Watchdog to restart it. RE Manager can be restarted at any time: the restart
    should not affect executed plans or the state of the queue or RE Worker. Response to this
    command is never returned, so it can also be used to test how the system handles communication
    timeouts. It takes 5 seconds of RE Manager inactivity befor it is restarted. The following cases
    are tested:
    - RE Manager is killed and restarted before queue processing is started.
    - RE Manager is killed and restarted while the 1st plan in the queue is executed.
    - RE Manager is killed while the 1st plan is still executed and is not restarted before
    the plan execution is finished. RE Manager is supposed to recognize that the plan is completed,
    process the report and start processing of the next plan.
    - RE Manager is killed and restarted while the 1st plan is in 'paused' state. RE Manager is
    supposed to switch to 'paused' state at the restart. The plan can execution can be resumed.
    """
    assert wait_for_condition(
        time=3, condition=condition_manager_idle
    ), "Timeout while waiting for manager to initialize."

    # Clear queue
    assert subprocess.call(["qserver", "queue", "clear"]) == SUCCESS

    plan = "{'name':'count', 'args':[['det1', 'det2']], 'kwargs':{'num': 10, 'delay': 1}}"
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS

    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    if time_kill == "before":
        # The command that kills manager always times out
        assert subprocess.call(["qserver", "manager", "kill", "test"]) == COM_ERROR
        ttime.sleep(8)  # It takes 5 seconds before the manager is restarted

        status = get_queue_state()
        assert status["manager_state"] == "idle"

    # Start queue processing
    assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS

    if isinstance(time_kill, int):
        ttime.sleep(time_kill)
        # The command that kills manager always times out
        assert subprocess.call(["qserver", "manager", "kill", "test"]) == COM_ERROR
        ttime.sleep(8)  # It takes 5 seconds before the manager is restarted

        status = get_queue_state()
        assert status["manager_state"] == "executing_queue"

    elif time_kill == "paused":
        ttime.sleep(3)
        assert subprocess.call(["qserver", "re", "pause", "deferred"]) == 0
        assert wait_for_condition(time=3, condition=condition_manager_paused)
        assert subprocess.call(["qserver", "manager", "kill", "test"]) == COM_ERROR
        ttime.sleep(8)  # It takes 5 seconds before the manager is restarted

        status = get_queue_state()
        assert status["manager_state"] == "paused"

        assert subprocess.call(["qserver", "re", "resume"]) == SUCCESS

    assert wait_for_condition(
        time=60, condition=condition_queue_processing_finished
    ), "Timeout while waiting for process to finish"

    n_plans, is_plan_running, n_history = get_reduced_state_info()
    assert n_plans == 0, "Incorrect number of plans in the queue"
    assert is_plan_running is False
    assert n_history == 2

    assert subprocess.call(["qserver", "environment", "close"]) == SUCCESS
    assert wait_for_condition(
        time=5, condition=condition_environment_closed
    ), "Timeout while waiting for environment to be closed"


# fmt: off
@pytest.mark.parametrize("additional_code, success", [
    # Nothing is added. Load profiles as usual.
    ("""
""", True),

    # Simulate profile that takes long time to load.
    ("""
\n
import time as ttime
ttime.sleep(20)

""", True),

    # Raise exception while loading the profile. This should cause RE Worker to exit.
    ("""
\n
raise Exception("This exception is raised to test if error handling works correctly")

""", False),
])
# fmt: on
def test_qserver_env_open_various_cases(re_manager_pc_copy, additional_code, success):  # noqa: F811

    _, pc_path = re_manager_pc_copy

    # Patch one of the startup files.
    patch_first_startup_file(pc_path, additional_code)

    # Wait until RE Manager is started
    assert wait_for_condition(time=10, condition=condition_manager_idle)

    # Attempt to create the environment. Long timeout.
    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
    assert wait_for_condition(time=30, condition=condition_manager_idle)

    status = get_queue_state()
    assert status["worker_environment_exists"] == success

    if not success:
        # Remove the offending patch and try to start the environment again. It should work
        patch_first_startup_file_undo(pc_path)
        assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
        assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    # Run a plan to make sure RE Manager is functional after the startup.
    plan = "{'name':'count', 'args':[['det1', 'det2']], 'kwargs':{'num': 10, 'delay': 1}}"
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS

    # Start queue processing
    assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
    ttime.sleep(2)
    status = get_queue_state()
    assert status["manager_state"] == "executing_queue"

    assert wait_for_condition(time=60, condition=condition_queue_processing_finished)
    n_plans, is_plan_running, n_history = get_reduced_state_info()
    assert n_plans == 0, "Incorrect number of plans in the queue"
    assert is_plan_running is False
    assert n_history == 1

    assert subprocess.call(["qserver", "environment", "close"]) == SUCCESS
    assert wait_for_condition(time=5, condition=condition_environment_closed)


# fmt: off
@pytest.mark.parametrize("option", [None, "on", "off"])
# fmt: on
def test_qserver_manager_stop_1(re_manager, option):  # noqa: F811
    """
    Method ``manager_stop``. Environment is in 'idle' state.
    """
    # Wait until RE Manager is started
    assert wait_for_condition(time=10, condition=condition_manager_idle)

    # Attempt to create the environment
    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
    assert wait_for_condition(time=timeout_env_open, condition=condition_manager_idle)

    cmd = ["qserver", "manager", "stop"]
    if option:
        cmd += ["safe", option]

    assert subprocess.call(cmd) == SUCCESS

    # Check if RE Manager was stopped.
    assert re_manager.check_if_stopped() is True


# fmt: off
@pytest.mark.parametrize("option", [None, "on", "off"])
# fmt: on
def test_qserver_manager_stop_2(re_manager, option):  # noqa: F811
    """
    Method ``manager_stop``. Environment is running a plan.
    """
    # Wait until RE Manager is started
    assert wait_for_condition(time=10, condition=condition_manager_idle)

    # Attempt to create the environment
    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
    assert wait_for_condition(time=timeout_env_open, condition=condition_manager_idle)

    plan = "{'name':'count', 'args':[['det1', 'det2']], 'kwargs':{'num': 10, 'delay': 1}}"
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS

    assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
    ttime.sleep(2)
    status = get_queue_state()
    assert status["manager_state"] == "executing_queue"

    cmd = ["qserver", "manager", "stop"]
    if option:
        cmd += ["safe", option]

    if option == "off":
        assert subprocess.call(cmd) == SUCCESS

        # Check if RE Manager was stopped.
        assert re_manager.check_if_stopped() is True

    else:
        assert subprocess.call(cmd) == REQ_FAILED

        assert wait_for_condition(time=60, condition=condition_queue_processing_finished)
        n_plans, is_plan_running, n_history = get_reduced_state_info()
        assert n_plans == 0, "Incorrect number of plans in the queue"
        assert is_plan_running is False
        assert n_history == 2


def test_queue_mode_set_1(re_manager):  # noqa F811
    """
    Basic test for ``qserver queue mode set`` command
    """
    assert subprocess.call(["qserver", "queue", "mode", "set", "loop", "True"]) == SUCCESS
    status = get_queue_state()
    assert status["plan_queue_mode"]["loop"] is True

    assert subprocess.call(["qserver", "queue", "mode", "set", "loop", "False"]) == SUCCESS
    status = get_queue_state()
    assert status["plan_queue_mode"]["loop"] is False


# fmt: off
@pytest.mark.parametrize("plist, exit_code", [
    (("set", "loop", "True"), SUCCESS),
    (("set",), SUCCESS),  # This should also work (no parameters -> the mode is not changed)
    (("unknown_option",), PARAM_ERROR),
    (("set", "loop"), PARAM_ERROR),  # Incorrect number of parameters
    (("set", "unknown_param", "True"), REQ_FAILED),  # Unsupported parameter name
    (("set", "loop", "true"), REQ_FAILED),  # Invalid parameter value
    (("set", "loop", "10"), REQ_FAILED),  # Invalid parameter value
])
# fmt: on
def test_queue_mode_set_2_fail(re_manager, plist, exit_code):  # noqa F811
    """
    Failing cases of the ``qserver queue mode set`` command
    """
    assert subprocess.call(["qserver", "queue", "mode", *plist]) == exit_code, str(plist)


# fmt: off
@pytest.mark.parametrize("pos, pos_result, success", [
    (None, 2, True),
    ("back", 2, True),
    ("front", 0, True),
    ("some", None, False),
    (0, 0, True),
    (1, 1, True),
    (2, 2, True),
    (3, 2, True),
    (100, 2, True),
    (-1, 1, True),
    (-2, 0, True),
    (-3, 0, True),
    (-100, 0, True),
])
# fmt: on
def test_queue_item_add_1(re_manager, pos, pos_result, success):  # noqa F811

    # Wait until RE Manager is started
    assert wait_for_condition(time=10, condition=condition_manager_idle)

    plan1 = "{'name':'count', 'args':[['det1']]}"
    plan2 = "{'name':'count', 'args':[['det1', 'det2']]}"

    # Create the queue with 2 entries
    assert subprocess.call(["qserver", "queue", "add", "plan", plan1]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "add", "plan", plan1]) == SUCCESS

    # Add another entry at the specified position
    params = [plan2]
    if pos is not None:
        params.insert(0, str(pos))

    res = subprocess.call(["qserver", "queue", "add", "plan", *params])
    if success:
        assert res == SUCCESS
    else:
        assert res == PARAM_ERROR

    resp = get_queue()
    assert len(resp["items"]) == (3 if success else 2)

    if success:
        assert resp["items"][pos_result]["args"] == [["det1", "det2"]]
        assert "item_uid" in resp["items"][pos_result]


def test_queue_item_add_2(re_manager):  # noqa F811
    """
    Failing cases: adding the plans that are expected to fail validation.
    """
    # Wait until RE Manager is started
    assert wait_for_condition(time=10, condition=condition_manager_idle)

    # Unknown plan name
    plan1 = "{'name':'count_test', 'args':[['det1']]}"
    # Unknown kwarg
    plan2 = "{'name':'count', 'args':[['det1', 'det2']], 'kwargs':{'abc': 10}}"

    # Both calls are expected to fail
    assert subprocess.call(["qserver", "queue", "add", "plan", plan1]) == REQ_FAILED
    assert subprocess.call(["qserver", "queue", "add", "plan", plan2]) == REQ_FAILED


# fmt: off
@pytest.mark.parametrize("before, target_pos, result_order", [
    (True, 0, [2, 0, 1]),
    (False, 0, [0, 2, 1]),
    (True, 1, [0, 2, 1]),
    (False, 1, [0, 1, 2]),
])
# fmt: on
def test_queue_item_add_3(re_manager, before, target_pos, result_order):  # noqa F811
    """
    Insert an item before or after the element with a given UID
    """
    # Wait until RE Manager is started
    assert wait_for_condition(time=10, condition=condition_manager_idle)

    plan1 = "{'name':'count', 'args':[['det1']]}"
    plan2 = "{'name':'count', 'args':[['det1', 'det2']]}"
    plan3 = "{'name':'count', 'args':[['det2']]}"

    assert subprocess.call(["qserver", "queue", "add", "plan", plan1]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "add", "plan", plan2]) == SUCCESS

    # Read queue.
    queue_1 = get_queue()["items"]
    assert len(queue_1) == 2
    uids_1 = [_["item_uid"] for _ in queue_1]

    params = ["before" if before else "after", uids_1[target_pos], plan3]
    assert subprocess.call(["qserver", "queue", "add", "plan", *params]) == SUCCESS

    # Check if the element was inserted in the right plance
    queue_2 = get_queue()["items"]
    assert len(queue_2) == 3
    uids_2 = [_["item_uid"] for _ in queue_2]
    for n, uid in enumerate(uids_2):
        n_res = result_order[n]
        if (n_res < 2) and (uid != uids_1[n_res]):
            assert False, f"uids_1: {uids_1}, uids_2: {uids_2}, result_order: {result_order}"


def test_queue_item_add_4(re_manager):  # noqa F811
    """
    Add instruction to the queue
    """
    plan1 = "{'name':'count', 'args':[['det1']]}"
    plan2 = "{'name':'count', 'args':[['det1', 'det2']]}"
    instruction = "queue-stop"

    assert subprocess.call(["qserver", "queue", "add", "plan", plan1]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "add", "instruction", instruction]) == SUCCESS
    assert subprocess.call(["qserver", "queue", "add", "plan", plan2]) == SUCCESS

    queue_1 = get_queue()["items"]
    assert len(queue_1) == 3
    assert queue_1[0]["item_type"] == "plan", str(queue_1[0])
    assert queue_1[1]["item_type"] == "instruction", str(queue_1[1])
    assert queue_1[2]["item_type"] == "plan", str(queue_1[2])


# fmt: off
@pytest.mark.parametrize("pos", [None, "back"])
# fmt: on
def test_queue_item_add_5_fail(re_manager, pos):  # noqa F811
    """
    No plan is supplied.
    """
    # Wait until RE Manager is started
    assert wait_for_condition(time=10, condition=condition_manager_idle)

    if pos:
        assert subprocess.call(["qserver", "queue", "add", "plan", pos]) == PARAM_ERROR
    else:
        assert subprocess.call(["qserver", "queue", "add", "plan"]) == PARAM_ERROR


# fmt: off
@pytest.mark.parametrize("pos", [10, "front", "back"])
# fmt: on
def test_queue_item_add_6_fail(re_manager, pos):  # noqa F811
    """
    Incorrect order of arguments (position is specified).
    """
    # Wait until RE Manager is started
    assert wait_for_condition(time=10, condition=condition_manager_idle)

    pos, plan = 10, "{'name':'count', 'args':[['det1']]}"
    params = [plan, str(pos)]
    assert subprocess.call(["qserver", "queue", "add", "plan", *params]) == PARAM_ERROR


# fmt: off
@pytest.mark.parametrize("params, exit_code", [
    # Error while processing message by the manager
    (["before_uid", "some_uid", "plan"], PARAM_ERROR),
    # Unknown keyword
    (["unknown_keyword", "some_uid", "plan"], PARAM_ERROR),
    # Incorrect order of arguments
    (["plan", "before_uid", "some_uid"], PARAM_ERROR),
    (["some_uid", "before_uid", "plan"], PARAM_ERROR),
    (["some_uid", "plan", "before_uid"], PARAM_ERROR),
])
# fmt: on
def test_queue_item_add_7_fail(re_manager, params, exit_code):  # noqa F811
    """
    Incorrect order of arguments (position is specified).
    """
    # Wait until RE Manager is started
    assert wait_for_condition(time=10, condition=condition_manager_idle)

    plan = "{'name':'count', 'args':[['det1']]}"
    params = [_ if _ != "plan" else plan for _ in params]
    assert subprocess.call(["qserver", "queue", "add", "plan", *params]) == exit_code


# fmt: on
@pytest.mark.parametrize("replace", [False, True])
@pytest.mark.parametrize("item_type", ["plan", "instruction"])
# fmt: off
def test_queue_item_update_1(re_manager, replace, item_type):  # noqa F811
    """
    Basic test for `queue_item_update` method.
    """
    plan1 = "{'name':'count', 'args':[['det1', 'det2']]}"
    plan2 = "{'name':'count', 'args':[['det1', 'det2']]}"
    instruction = "queue-stop"

    assert subprocess.call(["qserver", "queue", "add", "plan", plan1]) == SUCCESS

    queue_1 = get_queue()["items"]
    assert len(queue_1) == 1
    item_1 = queue_1[0]
    uid_to_replace = item_1["item_uid"]

    if item_type == "plan":
        item = plan2
    elif item_type == "instruction":
        item = instruction
    else:
        assert False, f"Unsupported item type '{item_type}'"
    option = "replace" if replace else "update"

    assert subprocess.call(["qserver", "queue", option, item_type, uid_to_replace, item]) == SUCCESS

    queue_2 = get_queue()["items"]
    assert len(queue_2) == 1
    item_2 = queue_2[0]

    if replace:
        assert item_2["item_uid"] != item_1["item_uid"]
    else:
        assert item_2["item_uid"] == item_1["item_uid"]
    item_2["item_type"] == item_type


# fmt: on
@pytest.mark.parametrize("replace", [False, True])
@pytest.mark.parametrize("item_type", ["plan", "instruction"])
# fmt: off
def test_queue_item_update_2_fail(re_manager, replace, item_type):  # noqa F811
    """
    Failing cases for `queue_item_update`: no matching UID is found in the queue.
    """
    plan1 = "{'name':'count', 'args':[['det1', 'det2']]}"
    plan2 = "{'name':'count', 'args':[['det1', 'det2']]}"
    instruction = "queue-stop"

    assert subprocess.call(["qserver", "queue", "add", "plan", plan1]) == SUCCESS

    queue_1 = get_queue()["items"]
    assert len(queue_1) == 1
    uid_to_replace = "non-existent-UID"

    if item_type == "plan":
        item = plan2
    elif item_type == "instruction":
        item = instruction
    else:
        assert False, f"Unsupported item type '{item_type}'"
    option = "replace" if replace else "update"

    assert subprocess.call(["qserver", "queue", option, item_type, uid_to_replace, item]) == REQ_FAILED

    queue_2 = get_queue()["items"]
    assert queue_1 == queue_2


# fmt: off
@pytest.mark.parametrize("item_type, env_exists", [
    ("plan", True),
    ("instruction", True),
    ("plan", False),
    ("instruction", False),
])
# fmt: on
def test_qserver_item_execute_1(re_manager, item_type, env_exists):  # noqa: F811
    """
    Long test runs a series of CLI commands.
    """
    plan_1 = "{'name':'count', 'args':[['det1', 'det2']]}"

    assert wait_for_condition(
        time=3, condition=condition_manager_idle
    ), "Timeout while waiting for manager to initialize."

    if env_exists:
        assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
        assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    expected_result = SUCCESS if env_exists else REQ_FAILED
    expected_n_history = 1 if env_exists and (item_type == "plan") else 0
    if item_type == "plan":
        item = ["plan", plan_1]
    elif item_type == "instruction":
        item = ["instruction", "queue-stop"]
    else:
        raise ValueError(f"Unknown item type '{item_type}'")

    assert subprocess.call(["qserver", "queue", "execute", *item]) == expected_result

    if env_exists:
        assert wait_for_condition(
            time=10, condition=condition_queue_processing_finished
        ), "Timeout while waiting for process to finish"

        assert subprocess.call(["qserver", "environment", "close"]) == SUCCESS
        assert wait_for_condition(
            time=5, condition=condition_environment_closed
        ), "Timeout while waiting for environment to be closed"

    n_plans, is_plan_running, n_history = get_reduced_state_info()
    assert n_plans == 0
    assert is_plan_running is False
    assert n_history == expected_n_history


# fmt: off
@pytest.mark.parametrize("pos, uid_ind, pos_result, success", [
    (None, None, 2, True),
    ("back", None, 2, True),
    ("front", None, 0, True),
    ("some", None, None, False),
    (0, None, 0, True),
    (1, None, 1, True),
    (2, None, 2, True),
    (3, None, None, False),
    (100, None, None, False),
    (-1, None, 2, True),
    (-2, None, 1, True),
    (-3, None, 0, True),
    (-4, None, 0, False),
    (-100, None, 0, False),
    (None, 0, 0, True),
    (None, 1, 1, True),
    (None, 2, 2, True),
    (None, 3, 2, False),
])
# fmt: on
def test_queue_item_get_remove(re_manager, pos, uid_ind, pos_result, success):  # noqa F811
    """
    Tests for ``queue_item_get`` and ``queue_item_remove`` requests.
    """
    plans = [
        "{'name':'count', 'args':[['det1']]}",
        "{'name':'count', 'args':[['det2']]}",
        "{'name':'count', 'args':[['det1', 'det2']]}",
    ]
    plans_args = [[["det1"]], [["det2"]], [["det1", "det2"]]]

    for plan in plans:
        assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS

    queue_1 = get_queue()["items"]
    assert len(queue_1) == 3
    uids_1 = [_["item_uid"] for _ in queue_1]
    uids_1.append("unknown_uid")  # Extra element (for one of the tests)

    if uid_ind is None:
        # Remove entry at the specified position
        args = [str(pos)] if (pos is not None) else []
    else:
        uid = uids_1[uid_ind]
        args = [uid]

    # Testing 'queue_item_get'. ONLY THE RETURN CODE IS TESTED.
    res = subprocess.call(["qserver", "queue", "item", "get", *args])
    if success:
        assert res == SUCCESS
    else:
        assert res == REQ_FAILED

    # Testing 'queue_item_remove'.
    res = subprocess.call(["qserver", "queue", "item", "remove", *args])
    if success:
        assert res == SUCCESS
    else:
        assert res == REQ_FAILED

    queue_2 = get_queue()["items"]
    assert len(queue_2) == (2 if success else 3)
    if success:
        ind = [0, 1, 2]
        ind.pop(pos_result)
        # Check that the right entry disappeared from the queue.
        assert queue_2[0]["args"] == plans_args[ind[0]]
        assert queue_2[1]["args"] == plans_args[ind[1]]


# fmt: off
@pytest.mark.parametrize("params, result_order, exit_code", [
    # 'params': positions are always represented as str, all int's are UIDs.
    (["0", "1"], [1, 0, 2], SUCCESS),
    (["2", "0"], [2, 0, 1], SUCCESS),
    (["2", "-3"], [2, 0, 1], SUCCESS),
    (["-1", "-3"], [2, 0, 1], SUCCESS),
    (["2", "-5"], [0, 1, 2], REQ_FAILED),  # Destination index out of range
    (["1", "3"], [0, 1, 2], REQ_FAILED),  # Destination index out of range
    (["front", "back"], [1, 2, 0], SUCCESS),
    (["back", "front"], [2, 0, 1], SUCCESS),
    ([1, "before", 0], [1, 0, 2], SUCCESS),
    ([0, "after", 1], [1, 0, 2], SUCCESS),
    (["1", "before", 0], [1, 0, 2], SUCCESS),  # Mixed pos->uid
    (["0", "after", 1], [1, 0, 2], SUCCESS),  # Mixed pos->uid
    ([1, "0"], [1, 0, 2], SUCCESS),  # Mixed uid->pos
    ([1, "2"], [0, 2, 1], SUCCESS),  # Mixed uid->pos
    (["1", "unknown_kwd", 0], [0, 1, 2], PARAM_ERROR),  # Mixed pos->uid
    (["0", "after"], [0, 1, 2], PARAM_ERROR),  # Second parameter 'after' is a keyword, UID is expected
    (["0"], [0, 1, 2], PARAM_ERROR),  # Not enough parameters
])
# fmt: on
def test_queue_item_get_move(re_manager, params, result_order, exit_code):  # noqa F811
    """
    Tests for ``queue_item_get`` and ``queue_item_remove`` requests.
    """
    plans = [
        "{'name':'count', 'args':[['det1']]}",
        "{'name':'count', 'args':[['det2']]}",
        "{'name':'count', 'args':[['det1', 'det2']]}",
    ]

    for plan in plans:
        assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == SUCCESS

    queue_1 = get_queue()["items"]
    assert len(queue_1) == 3
    uids_1 = [_["item_uid"] for _ in queue_1]
    uids_1.append("unknown_uid")  # Extra element (for one of the tests)

    # Replace ints with UIDs (positions are represented as strings)
    params = params.copy()
    for n, p in enumerate(params):
        if isinstance(p, int):
            params[n] = uids_1[p]

    # Testing 'queue_item_move'.
    assert subprocess.call(["qserver", "queue", "item", "move", *params]) == exit_code

    queue_2 = get_queue()["items"]
    assert len(queue_2) == 3
    uids_2 = [_["item_uid"] for _ in queue_2]

    # Compare the order of UIDs before and after moving the element
    uids_1_reordered = [uids_1[_] for _ in result_order]
    assert uids_1_reordered == uids_2


# fmt: off
@pytest.mark.parametrize("deactivate", [False, True])
# fmt: on
def test_qserver_queue_stop(re_manager, deactivate):  # noqa: F811
    """
    Methods ``queue_stop`` and ``queue_stop_cancel``.
    """
    # Wait until RE Manager is started
    assert wait_for_condition(time=10, condition=condition_manager_idle)

    # Attempt to create the environment
    assert subprocess.call(["qserver", "environment", "open"]) == 0
    assert wait_for_condition(time=timeout_env_open, condition=condition_manager_idle)

    plan = "{'name':'count', 'args':[['det1', 'det2']], 'kwargs':{'num': 10, 'delay': 1}}"
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == 0
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == 0

    # Queue is not running, so the request is expected to fail
    assert subprocess.call(["qserver", "queue", "stop"]) != 0
    status = get_queue_state()
    assert status["queue_stop_pending"] is False

    assert subprocess.call(["qserver", "queue", "start"]) == 0
    ttime.sleep(2)
    status = get_queue_state()
    assert status["manager_state"] == "executing_queue"

    assert subprocess.call(["qserver", "queue", "stop"]) == 0
    status = get_queue_state()
    assert status["queue_stop_pending"] is True

    if deactivate:
        ttime.sleep(1)
        assert subprocess.call(["qserver", "queue", "stop", "cancel"]) == 0
        status = get_queue_state()
        assert status["queue_stop_pending"] is False

    assert wait_for_condition(time=60, condition=condition_manager_idle)
    n_plans, is_plan_running, n_history = get_reduced_state_info()
    assert n_plans == (0 if deactivate else 1)
    assert is_plan_running is False
    assert n_history == (2 if deactivate else 1)
    status = get_queue_state()
    assert status["queue_stop_pending"] is False


def test_qserver_ping(re_manager):  # noqa: F811
    """
    Methods ``ping``: basic test
    """
    # Wait until RE Manager is started
    assert wait_for_condition(time=10, condition=condition_manager_idle)

    # Send 'ping' request
    assert subprocess.call(["qserver", "ping"]) == 0


# fmt: off
@pytest.mark.parametrize("option, exit_code", [
    (None, SUCCESS),
    ("active", SUCCESS),
    ("open", SUCCESS),
    ("closed", SUCCESS),
    ("some_unknown", PARAM_ERROR)
])
# fmt: on
def test_qserver_re_runs(re_manager, option, exit_code):  # noqa: F811
    """
    Basic test for ``re_runs`` method. There is no easy way to verify if the response
    was correct, so we just check if all supported combinations of parameters are accepted.
    """
    # Wait until RE Manager is started
    assert wait_for_condition(time=10, condition=condition_manager_idle)

    params = [option] if option else []
    assert subprocess.call(["qserver", "re", "runs", *params]) == exit_code


_script_to_upload_1 = """
def count_modified(detectors, num=1, delay=None):
    yield from count(detectors=detectors, num=num, delay=delay)
"""


# fmt: off
@pytest.mark.parametrize("run_in_background", [False, True])
@pytest.mark.parametrize("update_re", [False, True])
# fmt: on
def test_script_upload_1(re_manager, tmp_path, run_in_background, update_re):  # noqa: F811
    """
    Tests for 'qserver script upload'. The uploaded script does not change RE, so the tests
    simply checks if 'update-re' parameter is accepted.
    """
    # Create a file with the script
    script_fln = "script_to_upload.py"
    script_path = os.path.join(tmp_path, script_fln)
    with open(script_path, "w") as f:
        f.writelines(_script_to_upload_1)

    params = ["qserver", "script", "upload", script_path]
    if run_in_background:
        params.append("background")
    if update_re:
        params.append("update-re")

    # Call is expected to fail (environment is not open)
    assert subprocess.call(params) == REQ_FAILED

    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    # Check that 'count_modified' plan does not exist before the script is uploaded
    resp1, _ = zmq_single_request("plans_existing")
    assert resp1["success"] is True
    assert "count_modified" not in resp1["plans_existing"]

    if run_in_background:
        # Start some foreground process (run a function)
        item = r"{'name': 'function_sleep', 'kwargs': {'time': 2}}"
        params_fg = ["qserver", "function", "execute", item]
        assert subprocess.call(params_fg) == SUCCESS

    assert subprocess.call(params) == SUCCESS
    if run_in_background:
        ttime.sleep(2)  # Wait for the task to be completed
    assert wait_for_condition(time=5, condition=condition_manager_idle)

    # Check that 'count_modified' function was loaded into the environment
    resp2, _ = zmq_single_request("plans_existing")
    assert resp2["success"] is True
    assert "count_modified" in resp2["plans_existing"]

    assert subprocess.call(["qserver", "environment", "close"]) == SUCCESS
    assert wait_for_condition(time=5, condition=condition_environment_closed)


def test_script_upload_2_fail(re_manager, tmp_path):  # noqa: F811
    """
    Tests for 'qserver script upload': failing cases.
    """
    script_path = os.path.join(tmp_path, "script.py")
    with open(script_path, "w") as f:
        f.writelines(_script_to_upload_1)

    # File does not exist (IO error)
    assert subprocess.call(["qserver", "script", "upload", os.path.join(tmp_path, "script2.py")]) == PARAM_ERROR

    # Invalid parameter
    assert subprocess.call(["qserver", "script", "upload", script_path, "invalid_param"]) == PARAM_ERROR


# fmt: off
@pytest.mark.parametrize("run_in_background", [False, True])
# fmt: on
def test_function_execute_1(re_manager, run_in_background):  # noqa: F811
    """
    Tests for 'qserver function execute'.
    """
    item = r"{'name': 'function_sleep', 'kwargs': {'time': 2}}"
    params = ["qserver", "function", "execute", item]
    params_fg = params.copy()  # Parameters for foreground task
    if run_in_background:
        params.append("background")

    # Call is expected to fail (environment is not open)
    assert subprocess.call(params) == REQ_FAILED

    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    if run_in_background:
        # Start a foreground task (the same function) to test if a function can be run in the background
        assert subprocess.call(params_fg) == SUCCESS
        # Attempt to start another foreground process
        assert subprocess.call(params_fg) == REQ_FAILED

    assert subprocess.call(params) == SUCCESS
    if run_in_background:
        ttime.sleep(3)  # Wait for the task to be completed
    assert wait_for_condition(time=5, condition=condition_manager_idle)

    assert subprocess.call(["qserver", "environment", "close"]) == SUCCESS
    assert wait_for_condition(time=5, condition=condition_environment_closed)


def test_function_execute_2_fail(re_manager):  # noqa: F811
    """
    Tests for 'qserver function execute': failing cases.
    """
    item = r"{'name': 'function_sleep', 'kwargs': {'time': 2}}"

    # Invalid parameter
    assert subprocess.call(["qserver", "function", "execute", item, "invalid_param"]) == PARAM_ERROR


def test_task_result_1(re_manager):  # noqa: F811
    """
    Tests for 'qserver task result_get'.
    """
    # The request should be successful for any 'task_uid'.
    task_uid = "01e80342-5e36-44de-bc86-9bd8d57c9885"
    assert subprocess.call(["qserver", "task", "result", task_uid]) == SUCCESS

    # Some cases of invalid parameters
    assert subprocess.call(["qserver", "task", "result"]) == PARAM_ERROR
    assert subprocess.call(["qserver", "task", "something"]) == PARAM_ERROR
    assert subprocess.call(["qserver", "task", "result", task_uid, "extra_param"]) == PARAM_ERROR


_sample_trivial_plan1 = """
def trivial_plan_for_unit_test():
    '''
    Trivial plan for unit test.
    '''
    yield from scan([det1, det2], motor, -1, 1, 10)
"""


# fmt: off
@pytest.mark.parametrize("reload_plans_devices", [True, False])
# fmt: on
def test_qserver_permissions_reload_1(re_manager_pc_copy, tmp_path, reload_plans_devices):  # noqa F811
    """
    Tests for ``qserver permissions reload [lists]`` API.
    """
    # pc_path = copy_default_profile_collection(tmp_path)
    _, pc_path = re_manager_pc_copy
    append_code_to_last_startup_file(pc_path, additional_code=_sample_trivial_plan1)

    # Generate the new list of allowed plans and devices and reload them
    gen_list_of_plans_and_devices(startup_dir=pc_path, file_dir=pc_path, overwrite=True)

    plan = "{'name': 'trivial_plan_for_unit_test'}"

    # Attempt to add the plan to the queue. The request is supposed to fail, because
    #   the initially loaded profile collection does not contain the plan.
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == REQ_FAILED

    # Reload profile collection
    params = ["permissions", "reload"]
    if reload_plans_devices:
        params.append("lists")
    assert subprocess.call(["qserver", *params]) == SUCCESS

    # Attempt to add the plan to the queue. It should be successful now.
    res_expected = SUCCESS if reload_plans_devices else REQ_FAILED
    assert subprocess.call(["qserver", "queue", "add", "plan", plan]) == res_expected


_permissions_dict_not_allow_count = {
    "user_groups": {
        "root": {"allowed_plans": [None], "allowed_devices": [None]},
        "admin": {"allowed_plans": [None], "forbidden_plans": ["^count$"], "allowed_devices": [None]},
    }
}


def test_qserver_permissions_set_get_1(re_manager, tmp_path):  # noqa F811
    """
    Tests for ``qserver permissions set`` and ``qserver permissions get``: basic tests.
    """

    user_group_permissions = _permissions_dict_not_allow_count

    # Create yaml file with user group permissions
    fl_path = os.path.join(tmp_path, "ugp.yaml")
    with open(fl_path, "w") as f:
        f.writelines(yaml.dump(user_group_permissions))

    # Set user group permissions
    assert subprocess.call(["qserver", "permissions", "set", fl_path]) == SUCCESS

    # Check that permissions were really changed
    resp1, _ = zmq_single_request("permissions_get")
    assert resp1["success"] is True, f"resp={resp1}"
    assert resp1["user_group_permissions"] == user_group_permissions

    # Check if 'qserver permissions get' works
    assert subprocess.call(["qserver", "permissions", "get"]) == SUCCESS


def test_qserver_permissions_set_get_2_fail(re_manager, tmp_path):  # noqa F811
    """
    Tests for ``qserver permissions set`` and ``qserver permissions get``: failing cases.
    """
    # Create yaml file with user group permissions
    fl_path = os.path.join(tmp_path, "ugp.yaml")
    with open(fl_path, "w") as f:
        f.writelines("This is not a valid permissions dictionary")  # Invalid file

    # Non-existing file
    assert subprocess.call(["qserver", "permissions", "set", os.path.join(tmp_path, "some_file")]) == PARAM_ERROR
    # No file name
    assert subprocess.call(["qserver", "permissions", "set"]) == PARAM_ERROR
    # Extra parameters
    assert subprocess.call(["qserver", "permissions", "set", fl_path, "extra_parameter"]) == PARAM_ERROR
    # Request is correct, but the file does not contain valid dictionary
    assert subprocess.call(["qserver", "permissions", "set", fl_path]) == REQ_FAILED

    # Extra parameters
    assert subprocess.call(["qserver", "permissions", "get", "extra_parameter"]) == PARAM_ERROR


# fmt: off
@pytest.mark.parametrize("test_mode", ["none", "ev"])
# fmt: on
def test_qserver_secure_1(monkeypatch, re_manager_cmd, test_mode):  # noqa: F811
    """
    Test operation of `qserver` CLI tool with enabled encryption. Test options to
    set the private key used by `qserver` using the environment variable.
    """
    public_key, private_key = generate_new_zmq_key_pair()

    if test_mode == "none":
        pass
    elif test_mode == "ev":
        # Set server public key (for 'qserver') using environment variable
        monkeypatch.setenv("QSERVER_ZMQ_PUBLIC_KEY", public_key)
        # Set private key for RE manager
        monkeypatch.setenv("QSERVER_ZMQ_PRIVATE_KEY", private_key)
        # Set public key used by test helper functions such as 'wait_for_condition'
        set_qserver_zmq_public_key(monkeypatch, server_public_key=public_key)
    else:
        raise RuntimeError(f"Unrecognized test mode '{test_mode}'")

    # Security enabled by setting
    re_manager_cmd([])

    _plan1 = '{"name": "count", "args": [["det1", "det2"]], "item_type": "plan"}'
    _plan2 = '{"name": "scan", "args": [["det1", "det2"], "motor", -1, 1, 10], "item_type": "plan"}'

    # Add 2 plans
    assert subprocess.call(["qserver", "queue", "add", "plan", _plan1]) == 0
    assert subprocess.call(["qserver", "queue", "add", "plan", _plan2]) == 0

    # Request the list of allowed plans and devices (we don't check what is returned)
    assert subprocess.call(["qserver", "allowed", "plans"], stdout=subprocess.DEVNULL) == SUCCESS
    assert subprocess.call(["qserver", "allowed", "devices"], stdout=subprocess.DEVNULL) == SUCCESS

    # Request the list of existing plans and devices (we don't check what is returned)
    assert subprocess.call(["qserver", "existing", "plans"], stdout=subprocess.DEVNULL) == SUCCESS
    assert subprocess.call(["qserver", "existing", "devices"], stdout=subprocess.DEVNULL) == SUCCESS

    assert subprocess.call(["qserver", "environment", "open"]) == SUCCESS
    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    state = get_queue_state()
    assert state["items_in_queue"] == 2
    assert state["items_in_history"] == 0

    assert subprocess.call(["qserver", "queue", "start"]) == SUCCESS
    assert wait_for_condition(
        time=20, condition=condition_queue_processing_finished
    ), "Timeout while waiting for process to finish"

    state = get_queue_state()
    assert state["items_in_queue"] == 0
    assert state["items_in_history"] == 2

    assert subprocess.call(["qserver", "environment", "close"]) == SUCCESS
    assert wait_for_condition(
        time=5, condition=condition_environment_closed
    ), "Timeout while waiting for environment to be closed"


# ================================================================================
#                            qserver-zmq-keys


def test_qserver_zmq_keys():
    """
    Test for ``qserver-zmq-keys`` CLI
    """
    # Generate key pair
    assert subprocess.call(["qserver-zmq-keys"]) == SUCCESS

    # Generated public key based on private key - invalid key (exception)
    assert subprocess.call(["qserver-zmq-keys", "--zmq-private-key", "abc"]) == EXCEPTION_OCCURRED

    # Generated public key based on private key - success
    _, private_key = generate_new_zmq_key_pair()
    print(f"Private key used for the test: '{private_key}'")
    assert subprocess.call(["qserver-zmq-keys", "--zmq-private-key", private_key]) == SUCCESS



import functools
import inspect
import re
import os


class UserNamespace:
    """
    Keeps reference to the global user namespace. The object is used to emulate
    ``IPython.get_ipython().user_ns``. The attribute ``__use_ipython`` holds a
    flag that tells if ``IPython`` (namespace etc.) should be used.
    """

    def __init__(self):
        # Default: use IPython if executed from IPython.
        self.__user_ns = {}
        self.__use_ipython = True

    @property
    def user_ns(self):
        return self.__user_ns

    @user_ns.setter
    def user_ns(self, ns):
        raise RuntimeError("Attempting to set read-only property ``user_ns``")

    @property
    def use_ipython(self):
        return self.__use_ipython

    @use_ipython.setter
    def use_ipython(self, use_ipython):
        raise RuntimeError("Attempting to set read-only property ``use_ipython``")

    def set_user_namespace(self, *, user_ns, use_ipython=False):
        """
        Set reference to the user namespace. Typically the reference will point to
        a dictionary, but it also may be pointed to IPython ``user_ns``. The second
        parameter (``use_ipython``) indicates if IPython is used.

        Parameters
        ----------
        user_ns: dict
            Reference to the namespace.
        use_ipython: boolean
            Indicates if IPython is used. Typically it should be set ``False``.


        Raises
        ------
        TypeError
            Parameter ``user_ns`` is not a dictionary.
        """
        if not isinstance(user_ns, dict):
            raise TypeError(f"Parameter 'user_ns' must be of type 'dict': type(ns) is '{type(dict)}'")
        self.__user_ns = user_ns
        self.__use_ipython = bool(use_ipython)


global_user_namespace = UserNamespace()


def set_user_ns(func):
    """
    The decorator may be used to let the function access Run Engine namespace.
    Note, that the decorator may behave differently if used in start up files,
    because of patching of the ``get_ipython()`` function.

    If the function is executed from IPython, then the namespace is defined as
    ``IPython.get_ipython().user_ns``. If function is executed in python, then
    the namespace is passed as ``global_user_namespace.user_ns``. The object
    ``global_user_namespace`` is only a container for the storage, so its
    property ``user_ns`` should be set to reference RE namespace before the
    function is executed. Note, that the reference may be changed at any time -
    the namespace can be (and often is) different than the one used during
    import.

    The decorated function must have one kwargs: ``user_ns`` - reference
    to RE namespace. An optional kwarg ``ipython`` is a reference to IPython,
    which is None if the function called not from IPython.
    """

    def get_user_ns():
        from IPython import get_ipython

        ip = get_ipython()

        # This is needed for the case when the functions are executed from IPython,
        #   but we needed them to work exactly as if they were run without IPython.
        if not global_user_namespace.use_ipython:
            ip = None

        if ip:
            # Select IPython namespace (only if the code is run from IPython)
            user_ns = ip.user_ns
        else:
            # Select global user namespace that is manually set
            user_ns = global_user_namespace.user_ns
        return ip, user_ns

    # Parameter 'ipython' is optional
    is_ipython_in_sig = "ipython" in inspect.signature(func).parameters

    if inspect.isgeneratorfunction(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ip, user_ns = get_user_ns()
            kwargs.update({"user_ns": user_ns})
            if is_ipython_in_sig:
                kwargs.update({"ipython": ip})
            return (yield from func(*args, **kwargs))

    else:

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ip, user_ns = get_user_ns()
            kwargs.update({"user_ns": user_ns})
            if is_ipython_in_sig:
                kwargs.update({"ipython": ip})
            return func(*args, **kwargs)

    params_to_remove = ("user_ns", "ipython")
    sig_params = inspect.signature(wrapper).parameters
    sig = inspect.Signature([sig_params[_] for _ in sig_params if _ not in params_to_remove])
    wrapper.__signature__ = sig

    return wrapper


def load_devices_from_happi(device_names, *, namespace, **kwargs):
    """
    Load the devices from Happi based on the list of ``device_names``. The elements of the list
    may be strings (device name) or tuples of two strings (device name used for database search and
    the name with which the device is loaded in the script namespace). Two names may be needed
    in case the devices are stored in the database using compound names that contain beamline
    acronyms (e.g. ``abc_det`` and ``def_det`` for beamlines ABC and DEF) but they are expected
    to have the name ``det`` when loaded during ABC and DEF beamline startup respectively (see examples).

    The devices are loaded into a namespace referenced by ``namespace`` parameter.
    The function may be called multiple times in a row for the same namespace to populate
    it with results of multiple searches. The function also returns the list of names of loaded devices.

    Happi should be configured before the function could be used: Happi configuration file should be
    created and environment variable ``HAPPI_CFG`` with the path to the configuration file should be set.
    For example, if JSON Happi database is contained in the file ``path=/home/user/happi/database.json``
    and the path to configuration file is ``path=/home/user/happi.ini``, then the environment variable
    should be set as

    .. code-block::

        HAPPI_CFG=/home/user/happi.ini

    and configuration file ``happi.ini`` should contain

    .. code-block::

        [DEFAULT]
        backend=json
        path=/home/user/happi/database.json

    Examples
    --------

    .. code-block:: python

        # Load devices 'det1' and 'motor1'.
        load_devices_from_happi(["det1", "motor1"], namespace=locals())
        # Works exactly the same as the previous example. If the second tuple element is
        #   evaluated as boolean value of False, then it is ignored and the device is not renamed.
        load_devices_from_happi([("det1", ""), ("motor1", "")], namespace=locals())

        # Load 'abc_det1' as 'det1' and 'abc_motor1' as 'motor1'.
        load_devices_from_happi([("abc_det1", "det1"), ("abc_motor1", "motor1")], namespace=locals())

        # Get the dictionary of devices
        device_dict = {}
        load_devices_from_happi(["det1", "motor1"], namespace=device_dict)
        # 'device_dict': {"dev1": device1, "motor1": motor1}

        # Obtain the list of updated items (devices)
        item_list = load_devices_from_happi([("abc_det1", "det1"), ("abc_motor1", "motor1")], namespace=locals())
        # 'item_list': ["det1", "motor1"]

    Parameters
    ----------
    device_names : list(str) or list(tuple(str))
        List of device names. Elements of the list could be strings or tuples (lists) of string with
        two elements. If an element is a tuple of two names, then the first name is used in database
        search and the second name is the name of the device after it is loaded into the namespace.
        Each element of the list is processed separately and may contain mix of strings and tuples.
        If the second name in the tuple is an empty string, then it is ignored and
        the device is imported with the same name as used in database. It is expected that search
        will result in one found device per device name, otherwise ``RuntimeError`` is raised.
    namespace : dict
        Reference to the namespace where the devices should be loaded. It can be the reference
        to ``global()`` or ``locals()`` of the startup script or reference to a dictionary.
        It is a required keyword argument.
    kwargs : dict
        Additional search parameters.

    Returns
    -------
    list(str)
        the list of names of items in the ``namespace`` that were updated by the function.

    Raises
    ------
    RuntimeError
        Search for one of the device names returns no devices or more than one device.
    """

    kwargs = kwargs or {}

    if not isinstance(namespace, dict):
        raise TypeError(
            f"Parameter 'namespace' must be a dictionary: the value of type {type(namespace)} was passed instead"
        )

    # Verify that 'device_names' has correct type
    if not isinstance(device_names, (tuple, list)):
        raise TypeError(
            "Parameter 'device_names' value must be a tuple or a list: "
            f"type(device_names) = {type(device_names)}"
        )
    for n, name in enumerate(device_names):
        if not isinstance(name, (str, tuple, list)):
            raise TypeError(
                f"Parameter 'device_names': element #{n} must be str, tuple or list: " f"device_names[n] = {name}"
            )
        if isinstance(name, (tuple, list)):
            if len(name) != 2 or not isinstance(name[0], str) or not isinstance(name[1], str):
                raise TypeError(
                    f"Parameter 'device_names': element #{n} is expected to be in the form "
                    f"('name_in_db', 'name_in_namespace'): device_names[n] = {name}"
                )
            elif name[1] and not re.search(r"^[a-z][_a-z0-9]*$", name[1]):
                raise TypeError(
                    f"The device '{name[0]}' can not be renamed: The new device name '{name[1]}' "
                    "may consist of lowercase letters, numbers and '_' and must start from lowercase letter"
                )

    from happi import Client, load_devices

    client = Client.from_config()

    results = []
    for d_name in device_names:
        if isinstance(d_name, str):
            name_db, name_ns = d_name, None
        else:
            name_db, name_ns = d_name

        # Assemble search parameters
        search_params = dict(kwargs)
        search_params.update({"name": name_db})

        res = client.search(**search_params)
        if not res:
            raise RuntimeError(
                f"No devices with name '{name_db}' were found in Happi database. "
                f"Search parameters {search_params}"
            )
        elif len(res) > 1:
            raise RuntimeError(
                f"Multiple devices with name '{name_db}' were found in Happi database. "
                f"Search parameters {search_params}"
            )
        else:
            r = res[0]
            # Modify the object name (if needed)
            if name_ns:
                # In order for the following conversion to work properly, the name of
                #   the device should be specified only once (as `name` attribute),
                #   and aliases `{{name}}` should be used if the name is used as any other
                #   parameter. This is standard and recommended practice for instantiating
                #   Happi items (devices).
                #
                # In search results, the reference to the device is `r._device`.
                # `r.metadata` contains expanded metadata (it is probably not used, but it
                #   is a good idea to change it as well for consistency. We don't touch `_id`.
                # The modified data is not expected to be saved to the database.
                setattr(r._device, "name", name_ns)
                r.metadata["name"] = name_ns
            # Instantiate the object
            results.append(r)

    ns = load_devices(*[_.item for _ in results])
    ns_dict = ns.__dict__
    namespace.update(ns_dict)
    return list(ns_dict)


# Name of the environment variable may change in the future. Use API to set/clear the variable.
#   The variable name should not be included in the documentation.
_env_re_worker_active = "QSERVER_RE_WORKER_ACTIVE"


def set_re_worker_active():
    """
    Set the environment variable used to determine if the current process is RE Worker process.
    Subsequent calls to ``is_re_worker_active`` in the current process will return ``True``.
    THIS FUNCTION SHOULD NEVER BE CALLED IN STARTUP SCRIPTS.
    """
    os.environ[_env_re_worker_active] = "1"


def clear_re_worker_active():
    """
    Clear the environment variable used to determine if the current process is RE Worker process.
    Subsequent calls to ``is_re_worker_active`` in the current process will return ``False``.
    THIS FUNCTION SHOULD NEVER BE CALLED IN STARTUP SCRIPTS.
    """
    if _env_re_worker_active in os.environ:
        del os.environ[_env_re_worker_active]


def is_re_worker_active():
    """
    The function can be used in startup scripts or modules to check if the script is imported or
    executed in RE Worker environment. For example, an experimental plan may contain interactive
    features that should be disabled if the plan is executed remotely:

    .. code-block:: python

        from bluesky_queueserver import is_re_worker_active

        ...

        if is_re_worker_active():
            (code without interactive features, e.g. reading data from a file)
        else:
            (code with interactive features, e.g. manual data input)

        ...

    Returns
    -------
    boolean
        ``True`` - the code is executed in RE Worker environment, otherwise ``False``.
    """
    return os.environ.get(_env_re_worker_active, "false").lower() not in ("", "n", "no", "f", "false", "off", "0")



import pandas as pd
import os


def spreadsheet_to_plan_list(*, spreadsheet_file, data_type, file_name, **kwargs):
    """
    Process trivial spreadsheet with plan parameters (for use in unit tests).
    The spreadsheet is expected to contain parameters of 'count' plan in the form:
        name   num   delay
    0   count  5     1
    1   count  6     0.5

    NOTE: this function returns item (plan) parameters without setting 'item_type'.
    Those items are considered plans (equivalent to ``item_type`` set as ``plan``).
    The items in the list may also be instructions (or other item types when they
    are supported). In this case each item should include ``item_type`` set explicitly.

    Parameters
    ----------
    spreadsheet_file : file
        readable file object
    data_type : str
        string that represents type of data in the spreadsheets. May be used to
        distinguish between different types of spreadsheets used at the same beamline.
    file_name : str
        name of the uploaded spreadsheet file, may be used to determine file type by
        looking at the extension.

    Returns
    -------
    plan_list : list(dict)
        Dictionary representing a list of plans extracted from the spreadsheet.
    """

    # Test that data type is successfully passed. In real processing functions we would
    #   check if it matches one of the supported functions.
    if data_type == "unsupported":
        raise ValueError(f"Unsupported data type: '{data_type}'")

    # If custom processing function (such as this one) returns None, then the spreadsheet
    #   is passed to the default processing function.
    if data_type == "process_with_default_function":
        return None

    # Check that file name has extension 'xlsx'. Raise the exception otherwise.
    ext = os.path.splitext(file_name)[1]
    if ext != ".xlsx":
        raise ValueError(f"Unsupported file (extension '{ext}')")

    df = pd.read_excel(spreadsheet_file, index_col=0, engine="openpyxl")
    plan_list = []
    n_rows, _ = df.shape
    for nr in range(n_rows):
        plan = {
            "name": df["name"][nr],
            "args": [["det1", "det2"]],
            "kwargs": {"num": int(df["num"][nr]), "delay": float(df["delay"][nr])},
        }
        plan_list.append(plan)
    return plan_list



import os
import inspect
import pytest
import sys

from bluesky_queueserver.manager.profile_tools import (
    global_user_namespace,
    load_devices_from_happi,
    set_re_worker_active,
    clear_re_worker_active,
    is_re_worker_active,
)
from bluesky_queueserver.manager.profile_ops import load_profile_collection, gen_list_of_plans_and_devices

from ..comms import zmq_single_request

from .common import (
    wait_for_condition,
    condition_environment_created,
    condition_environment_closed,
    condition_manager_idle,
    copy_default_profile_collection,
    patch_first_startup_file,
)
from .common import re_manager_cmd  # noqa: F401

# User name and user group name used throughout most of the tests.
_user, _user_group = "Testing Script", "admin"


def create_local_imports_files(tmp_path):
    path_dir = os.path.join(tmp_path, "dir_local_imports")
    fln_func = os.path.join(path_dir, "file_func.py")
    fln_gen = os.path.join(path_dir, "file_gen.py")

    os.makedirs(path_dir, exist_ok=True)

    # Create file1
    code1 = """
from bluesky_queueserver.manager.profile_tools import set_user_ns

# Function that has the parameter 'ipython'
@set_user_ns
def f1(some_value, user_ns, ipython):
    user_ns["func_was_called"] = "func_was_called"
    return (some_value, user_ns["v_from_namespace"], bool(ipython))

# Function that has no parameter 'ipython'
@set_user_ns
def f1a(some_value, user_ns):
    user_ns["func_A_was_called"] = "func_was_called"
    return (some_value, user_ns["v_from_namespace"])

"""
    with open(fln_func, "w") as f:
        f.writelines(code1)

    # Create file2
    code2 = """
from bluesky_queueserver.manager.profile_tools import set_user_ns

# Function that has the parameter 'ipython'
@set_user_ns
def f2(some_value, user_ns, ipython):
    user_ns["gen_was_called"] = "gen_was_called"
    yield (some_value, user_ns["v_from_namespace"], bool(ipython))

# Function that has no parameter 'ipython'
@set_user_ns
def f2a(some_value, user_ns):
    user_ns["gen_A_was_called"] = "gen_was_called"
    yield (some_value, user_ns["v_from_namespace"])

@set_user_ns
def f3(some_value, user_ns, ipython):
    user_ns["value_f3"] = some_value

f3(91)

"""
    with open(fln_gen, "w") as f:
        f.writelines(code2)


patch_code = """
from dir_local_imports.file_func import f1, f1a
from dir_local_imports.file_gen import f2, f2a

from bluesky_queueserver.manager.profile_tools import set_user_ns

@set_user_ns
def f4(some_value, user_ns, ipython):
    user_ns["value_f4"] = some_value

f4(90)

"""


def test_set_user_ns_1(tmp_path):
    """
    Tests for ``set_user_ns`` decorator. The functionality of the decorator
    is fully tested (only without IPython):
    - using ``global_user_namespace`` to pass values in and out of the function
      defined in the imported module (emulation of ``get_ipython().user_ns``).
    - checking if the function is executed from IPython (only for the function
      defined in the imported module).
    """
    pc_path = copy_default_profile_collection(tmp_path)

    create_local_imports_files(pc_path)
    patch_first_startup_file(pc_path, patch_code)

    nspace = load_profile_collection(pc_path)
    assert len(nspace) > 0, "Failed to load the profile collection"
    assert "f1" in nspace, "Test for local imports failed"
    assert "f2" in nspace, "Test for local imports failed"

    # Test if the decorator `set_user_ns` does not change function type
    assert inspect.isgeneratorfunction(nspace["f1"]) is False
    assert inspect.isgeneratorfunction(nspace["f2"]) is True

    # Check if the extra arguments are removed from the function signature
    def check_signature(func):
        params = inspect.signature(func).parameters
        assert "user_ns" not in params
        assert "ipython" not in params

    check_signature(nspace["f1"])
    check_signature(nspace["f1a"])
    check_signature(nspace["f2"])
    check_signature(nspace["f2a"])

    assert nspace["value_f3"] == 91
    assert nspace["value_f4"] == 90

    # Test function
    global_user_namespace.set_user_namespace(user_ns=nspace, use_ipython=False)
    global_user_namespace.user_ns["v_from_namespace"] = "value-sent-to-func"
    assert nspace["v_from_namespace"] == "value-sent-to-func"

    result_func = nspace["f1"](60)
    assert nspace["func_was_called"] == "func_was_called"
    assert result_func[0] == 60
    assert result_func[1] == "value-sent-to-func"
    assert result_func[2] is False

    result_func = nspace["f1a"](65)
    assert nspace["func_A_was_called"] == "func_was_called"
    assert result_func[0] == 65
    assert result_func[1] == "value-sent-to-func"

    # Test generator
    global_user_namespace.user_ns["v_from_namespace"] = "value-sent-to-gen"
    result_func = list(nspace["f2"](110))[0]
    assert nspace["gen_was_called"] == "gen_was_called"
    assert result_func[0] == 110
    assert result_func[1] == "value-sent-to-gen"
    assert result_func[2] is False

    result_func = list(nspace["f2a"](115))[0]
    assert nspace["gen_A_was_called"] == "gen_was_called"
    assert result_func[0] == 115
    assert result_func[1] == "value-sent-to-gen"


def test_global_user_namespace():
    """
    Basic test for ``global_user_namespace``.
    """
    ns = {"ab": 1, "cd": 2}
    global_user_namespace.set_user_namespace(user_ns=ns)
    assert global_user_namespace.user_ns == ns
    assert global_user_namespace.use_ipython is False

    global_user_namespace.set_user_namespace(user_ns={}, use_ipython=True)
    assert global_user_namespace.user_ns == {}
    assert global_user_namespace.use_ipython is True

    global_user_namespace.set_user_namespace(user_ns=ns, use_ipython=False)
    assert global_user_namespace.user_ns == ns
    assert global_user_namespace.use_ipython is False


_happi_json_db_1 = """
{
  "det": {
    "_id": "det",
    "active": true,
    "args": [],
    "device_class": "ophyd.sim.DetWithCountTime",
    "documentation": null,
    "kwargs": {
      "name": "{{name}}"
    },
    "name": "det",
    "type": "OphydItem"
  },
  "motor": {
    "_id": "motor",
    "active": true,
    "args": [],
    "device_class": "ophyd.sim.SynAxisNoPosition",
    "documentation": null,
    "kwargs": {
      "name": "{{name}}"
    },
    "name": "motor",
    "type": "OphydItem"
  },
  "motor1": {
    "_id": "motor1",
    "active": true,
    "args": [],
    "device_class": "ophyd.sim.SynAxisNoHints",
    "documentation": null,
    "kwargs": {
      "name": "{{name}}"
    },
    "name": "motor1",
    "type": "OphydItem"
  },
  "tst_motor2": {
    "_id": "tst_motor2",
    "active": true,
    "args": [],
    "device_class": "ophyd.sim.SynAxisNoHints",
    "documentation": null,
    "kwargs": {
      "name": "{{name}}"
    },
    "name": "tst_motor2",
    "type": "OphydItem"
  },
  "motor3": {
    "_id": "motor3",
    "active": true,
    "args": [],
    "device_class": "ophyd.sim.SynAxis",
    "documentation": null,
    "kwargs": {
      "name": "{{name}}"
    },
    "name": "motor3",
    "type": "OphydItem"
  },
  "motor3_duplicate_error": {
    "_id": "motor3",
    "active": false,
    "args": [],
    "device_class": "ophyd.sim.SynAxis",
    "documentation": null,
    "kwargs": {
      "name": "{{name}}"
    },
    "name": "motor3",
    "type": "OphydItem"
  }
}
"""


def _configure_happi(tmp_path, monkeypatch, json_devices):
    path_json = os.path.join(tmp_path, "sim_devices.json")
    path_ini = os.path.join(tmp_path, "happi.ini")

    happi_ini_text = f"[DEFAULT]\nbackend=json\npath={path_json}"

    with open(path_ini, "w") as f:
        f.write(happi_ini_text)

    with open(path_json, "w") as f:
        f.write(json_devices)

    monkeypatch.setenv("HAPPI_CFG", path_ini)


# fmt: off
@pytest.mark.parametrize("device_names, loaded_names, kw_args, success, errmsg", [
    ([], [], {}, True, ""),  # No devices are loaded if the list of devices is empty
    (("det", "motor"), ("det", "motor"), {}, True, ""),
    (["det", "motor"], ("det", "motor"), {}, True, ""),
    ((("det", ""), ["motor", ""]), ("det", "motor"), {}, True, ""),
    (("det", ["motor", ""]), ("det", "motor"), {}, True, ""),
    (("det", ("motor", ""), ("tst_motor2", "motor2")), ("det", "motor", "motor2"), {}, True, ""),
    # This is not typical use case, but the same device may be loaded multiple times
    #   with different names if needed.
    ((("motor1", "motor1_copy1"), ("motor1", "motor1_copy2")), ("motor1_copy1", "motor1_copy2"), {}, True, ""),
    # Incorrect type of the device list
    (10, ("det", "motor"), {}, False, "Parameter 'device_names' value must be a tuple or a list"),
    ("string", ("det", "motor"), {}, False, "Parameter 'device_names' value must be a tuple or a list"),
    # Incorrecty type or form of a device list element
    (("det", 10), ("det", "motor"), {}, False, "Parameter 'device_names': element .* must be str, tuple or list"),
    ((10, "motor"), ("det", "motor"), {}, False,
     "Parameter 'device_names': element .* must be str, tuple or list"),
    (("det", (10, "motor2")), ("det", "motor"), {}, False, "element .* is expected to be in the form"),
    (("det", ("tst_motor2", 10)), ("det", "motor"), {}, False, "element .* is expected to be in the form"),
    (("det", ("tst_motor2", "motor2", 10)), ("det", "motor"), {}, False,
     "element .* is expected to be in the form"),
    # No device found
    (("det", "motor10"), ("det", "motor10"), {}, False, "No devices with name"),
    # Multiple devices found (search for "motor3" yields multile devices, this is database issue)
    (("det", "motor3"), ("det", "motor3"), {}, False, "Multiple devices with name"),
    # Use additional search parameters. (Two entries for "motor3" differ in the value of `active` field.
    #   A single entry for `det` has `active==True`.)
    (("det", "motor3"), ("det", "motor3"), {"active": True}, True, ""),
    (("det", "motor3"), ("det", "motor3"), {"active": False}, False,
     "No devices with name 'det' were found in Happi database."),
    (("motor3",), ("motor3",), {"active": False}, True, ""),
    # Verify that valid device names are accepted
    (("det", ["motor", "motor3_new"]), ("det", "motor3_new"), {}, True, ""),
    # Invalid new device name
    (("det", ["motor", "Motor"]), ("det", "motor"), {}, False, "may consist of lowercase letters, numbers"),
    (("det", ["motor", "moTor"]), ("det", "motor"), {}, False, "may consist of lowercase letters, numbers"),
    (("det", ["motor", "_motor"]), ("det", "motor"), {}, False, "may consist of lowercase letters, numbers"),
    (("det", ["motor", " motor"]), ("det", "motor"), {}, False, "may consist of lowercase letters, numbers"),
    (("det", ["motor", "motor "]), ("det", "motor"), {}, False, "may consist of lowercase letters, numbers"),
    (("det", ["motor", "motor new"]), ("det", "motor"), {}, False, "may consist of lowercase letters, numbers"),
    (("det", ["motor", "motor_$new"]), ("det", "motor"), {}, False, "may consist of lowercase letters, numbers"),
    (("det", ["motor", "2motor_$new"]), ("det", "motor"), {}, False, "may consist of lowercase letters, numbers"),
])
# fmt: on
def test_load_devices_from_happi_1(tmp_path, monkeypatch, device_names, loaded_names, kw_args, success, errmsg):
    """
    Tests for ``load_devices_from_happi``.
    """
    _configure_happi(tmp_path, monkeypatch, json_devices=_happi_json_db_1)

    # Load as a dictionary
    if success:
        ns = {}
        dlist = load_devices_from_happi(device_names, namespace=ns, **kw_args)
        assert len(ns) == len(loaded_names), str(ns)
        for d in loaded_names:
            assert d in ns
        assert set(dlist) == set(loaded_names)
    else:
        with pytest.raises(Exception, match=errmsg):
            ns = {}
            load_devices_from_happi(device_names, namespace=ns, **kw_args)

    # Load in local namespace
    def _test_loading(device_names, loaded_names):
        if success:
            load_devices_from_happi(device_names, namespace=locals(), **kw_args)
            for d in loaded_names:
                assert d in locals()
        else:
            with pytest.raises(Exception, match=errmsg):
                load_devices_from_happi(device_names, namespace=locals(), **kw_args)

    _test_loading(device_names=device_names, loaded_names=loaded_names)


def test_load_devices_from_happi_2_fail(tmp_path, monkeypatch):
    """
    Function ``load_devices_from_happi``: parameter ``namespace`` is required and must be of type ``dict``.
    """
    _configure_happi(tmp_path, monkeypatch, json_devices=_happi_json_db_1)

    # Missing 'namespace' parameter
    with pytest.raises(TypeError, match="missing 1 required keyword-only argument: 'namespace'"):
        load_devices_from_happi(["det", "motor"])

    # Incorrect type of 'namespace' parameter
    with pytest.raises(TypeError, match="Parameter 'namespace' must be a dictionary"):
        load_devices_from_happi(["det", "motor"], namespace=[1, 2, 3])


def test_is_re_worker_active_1(monkeypatch):  # noqa: F811
    """
    Basic tests for ``set_re_worker_active``, ``clear_re_worker_active`` and ``is_re_worker_active`` functions.
    """
    monkeypatch.setattr(os, "environ", os.environ.copy())
    assert is_re_worker_active() is False
    set_re_worker_active()
    assert is_re_worker_active() is True
    clear_re_worker_active()
    assert is_re_worker_active() is False


# Minimalistic user permissions sufficient to start RE Manager
_user_groups_text = r"""user_groups:
  root:  # The group includes all available plan and devices
    allowed_plans:
      - null  # Allow all
    forbidden_plans:
      - null  # Nothing is forbidden
    allowed_devices:
      - null  # Allow all
    forbidden_devices:
      - null  # Nothing is forbidden
  admin:  # The group includes beamline staff, includes all or most of the plans and devices
    allowed_plans:
      - null  # Allow all
    forbidden_plans:
      - null  # Nothing is forbidden
    allowed_devices:
      - null  # Allow all
    forbidden_devices:
      - null  # Nothing is forbidden
"""


_startup_script_1 = """
from ophyd.sim import det1, det2
from bluesky.plans import count

from bluesky_queueserver.manager.profile_tools import is_re_worker_active
from dir1.file1 import f1

# Executed during import
if not is_re_worker_active():
    raise Exception("Importing startup script: RE Worker is not detected as active")

def sim_plan_1():
    '''
    Simple plan for tests.
    '''
    if f1() is not True:
        raise Exception("sim_plan_1: Function 'f1' did not return correct result.")

    if not is_re_worker_active():
        raise Exception("sim_plan_1: RE Worker is not detected as active")

    yield from count([det1, det2])
"""


def create_local_imports_dir(pc_path):
    path1 = os.path.join(pc_path, "dir1")
    fln1 = os.path.join(path1, "file1.py")

    os.makedirs(path1, exist_ok=True)

    # Create file1
    code1 = """
from bluesky_queueserver.manager.profile_tools import is_re_worker_active

# Executed during import
if not is_re_worker_active():
    raise Exception("Importing module 'file1': RE Worker is not detected as active")

def f1():
    if not is_re_worker_active():
        raise Exception("Function f1: RE Worker is not detected as active")
    return True
"""
    with open(fln1, "w") as f:
        f.writelines(code1)


# fmt: off
@pytest.mark.parametrize("option", ["startup_dir", "script", "module"])
# fmt: on
def test_is_re_worker_active_2(re_manager_cmd, tmp_path, monkeypatch, option):  # noqa: F811
    """
    Test that ``is_re_worker_active()`` API is working as expected in startup scripts.
    The sample startup scripts are modified to call ``is_re_worker_active()`` during import
    and during execution of a plan. The script is also importing the local module, which
    is calling the API during import and in the body of a function that is called during
    the plan execution. Two modes of operation are tested: (1) loading the profile collection
    in the current process (``gen_list_of_plans_and_devices`` function); (2) loading
    of startup profile collection and execution of the plan in RE Worker environment.
    """
    monkeypatch.setattr(os, "environ", os.environ.copy())

    # Load first script
    script_dir = os.path.join(tmp_path, "script_dir1")
    script_path = os.path.join(script_dir, "startup_script.py")

    os.makedirs(script_dir, exist_ok=True)
    with open(script_path, "w") as f:
        f.write(_startup_script_1)

    create_local_imports_dir(script_dir)

    path_user_permissions = os.path.join(script_dir, "user_group_permissions.yaml")
    with open(path_user_permissions, "w") as f:
        f.writelines(_user_groups_text)

    path_existing_plans_and_devices = os.path.join(script_dir, "existing_plans_and_devices.yaml")

    # Make sure that 'is_re_worker_active()' works as expected when the list of plans and devices is
    #   created (profile collection is loaded in the current process).
    assert is_re_worker_active() is False
    if option == "startup_dir":
        gen_list_of_plans_and_devices(startup_dir=script_dir, file_dir=script_dir, overwrite=True)
    elif option == "script":
        gen_list_of_plans_and_devices(startup_script_path=script_path, file_dir=script_dir, overwrite=True)
    elif option == "module":
        # Temporarily add module to the search path
        sys_path = sys.path
        monkeypatch.setattr(sys, "path", [str(tmp_path)] + sys_path)
        gen_list_of_plans_and_devices(
            startup_module_name="script_dir1.startup_script", file_dir=script_dir, overwrite=True
        )
    else:
        assert False, f"Unknown option '{option}'"
    assert is_re_worker_active() is False  # Make sure that that 'active' state is cleared

    if option == "startup_dir":
        re_manager_cmd(["--startup-dir", script_dir])
    elif option == "script":
        re_manager_cmd(
            [
                "--startup-script",
                script_path,
                "--user-group-permissions",
                path_user_permissions,
                "--existing-plans-devices",
                path_existing_plans_and_devices,
            ]
        )
    elif option == "module":
        # The manager will still start, but the environment can not load, because
        #   the module 'script_dir1.startup_script' is not installed.
        re_manager_cmd(
            [
                "--startup-module",
                "script_dir1.startup_script",
                "--user-group-permissions",
                path_user_permissions,
                "--existing-plans-devices",
                path_existing_plans_and_devices,
            ]
        )
        # Since the environment can not be loaded, the rest of the test will not work.
        #   We successfully loaded the module in `gen_list_of_plans_and_devices()`, so
        #   let's consider it success and interrupt the test.
        return
    else:
        assert False, f"Unknown option '{option}'"

    # Open environment and execute plan 'sim_plan_1'
    resp1, _ = zmq_single_request("permissions_reload", {"reload_plans_devices": True})
    assert resp1["success"] is True, f"resp={resp1}"

    # Add plan to the queue
    params = {"item": {"name": "sim_plan_1", "item_type": "plan"}, "user": _user, "user_group": _user_group}
    resp2, _ = zmq_single_request("queue_item_add", params)
    assert resp2["success"] is True, f"resp={resp2}"

    # Open the environment
    resp3, _ = zmq_single_request("environment_open")
    assert resp3["success"] is True
    assert wait_for_condition(time=10, condition=condition_environment_created)

    # Make sure that the status is not set in the current process
    assert is_re_worker_active() is False

    # Make sure that the run is in the queue and the history is empty
    status, _ = zmq_single_request("status")
    assert status["items_in_queue"] == 1
    assert status["items_in_history"] == 0

    # Start the queue
    resp4, _ = zmq_single_request("queue_start")
    assert resp4["success"] is True

    assert wait_for_condition(time=5, condition=condition_manager_idle)

    # Make sure that the run was executed successfully
    status, _ = zmq_single_request("status")
    assert status["items_in_queue"] == 0
    assert status["items_in_history"] == 1

    # Make sure that plan was completed successfully
    resp5, _ = zmq_single_request("history_get")
    assert resp5["success"] is True
    history = resp5["items"]
    assert history[-1]["result"]["exit_status"] == "completed"

    # Close the environment
    resp6, _ = zmq_single_request("environment_close")
    assert resp6["success"] is True, f"resp={resp6}"
    assert wait_for_condition(time=5, condition=condition_environment_closed)



import asyncio
import time as ttime

from bluesky_queueserver.manager.task_results import TaskResults


def test_TaskResults_update_uid():
    """
    TaskResults: Test that task result UID is updated.
    """

    async def testing():
        tr = TaskResults()

        uid = tr.task_results_uid
        assert isinstance(uid, str)
        tr._update_task_results_uid()
        assert tr.task_results_uid != uid

    asyncio.run(testing())


def test_TaskResults_add_running_task():
    """
    TaskResults: tests for ``add_running_task``, ``clear_running_task``
    """

    async def testing():
        tr = TaskResults()

        assert tr._running_tasks == {}
        uid = tr.task_results_uid

        await tr.add_running_task(task_uid="abc")
        await tr.add_running_task(task_uid="def", payload={"some_value": 10})
        assert len(tr._running_tasks) == 2

        assert tr._running_tasks["abc"]["payload"] == {}
        assert isinstance(tr._running_tasks["abc"]["time"], float)
        assert tr._running_tasks["def"]["payload"] == {"some_value": 10}
        assert isinstance(tr._running_tasks["def"]["time"], float)

        await tr.clear_running_tasks()
        assert tr._running_tasks == {}

        # UID is not expected to change
        assert tr.task_results_uid == uid

    asyncio.run(testing())


def test_TaskResults_remove_running_task():
    """
    TaskResults: tests for ``remove_running_task``
    """

    async def testing():
        tr = TaskResults()

        assert tr._running_tasks == {}
        uid = tr.task_results_uid

        await tr.add_running_task(task_uid="abc")
        await tr.add_running_task(task_uid="def", payload={"some_value": 10})
        assert len(tr._running_tasks) == 2

        await tr.remove_running_task(task_uid="abc")

        assert len(tr._running_tasks) == 1
        assert tr._running_tasks["def"]["payload"] == {"some_value": 10}
        assert isinstance(tr._running_tasks["def"]["time"], float)

        # UID is not expected to change
        assert tr.task_results_uid == uid

    asyncio.run(testing())


def test_TaskResults_add_completed_task():
    """
    TaskResults: tests for ``add_running_task``, ``clear_running_task``
    """

    async def testing():
        tr = TaskResults()

        uid1 = tr.task_results_uid

        # Add running tasks. The running tasks should be removed as completed tasks
        #   with the same UID are added.
        await tr.add_running_task(task_uid="abc", payload={"some_value": "arbitrary_payload"})
        await tr.add_running_task(task_uid="def", payload={"some_value": "arbitrary_payload"})

        assert tr.task_results_uid == uid1

        assert tr._completed_tasks_time == []
        assert tr._completed_tasks_data == {}
        assert len(tr._running_tasks) == 2

        await tr.add_completed_task(task_uid="abc")
        uid2 = tr.task_results_uid
        await tr.add_completed_task(task_uid="def", payload={"some_value": 10})
        uid3 = tr.task_results_uid

        assert len(tr._completed_tasks_time) == 2
        assert len(tr._completed_tasks_data) == 2
        assert len(tr._running_tasks) == 0

        assert uid1 != uid2
        assert uid2 != uid3

        assert tr._completed_tasks_data["abc"]["payload"] == {}
        assert isinstance(tr._completed_tasks_data["abc"]["time"], float)
        assert tr._completed_tasks_time[0]["task_uid"] == "abc"
        assert tr._completed_tasks_time[0]["time"] == tr._completed_tasks_data["abc"]["time"]
        assert tr._completed_tasks_data["def"]["payload"] == {"some_value": 10}
        assert isinstance(tr._completed_tasks_data["def"]["time"], float)
        assert tr._completed_tasks_time[1]["task_uid"] == "def"
        assert tr._completed_tasks_time[1]["time"] == tr._completed_tasks_data["def"]["time"]

        await tr.clear()
        assert tr._completed_tasks_time == []
        assert tr._completed_tasks_data == {}

        # UID is not expected to change
        assert tr.task_results_uid == uid3

    asyncio.run(testing())


def test_TaskResults_clear():
    """
    TaskResults: tests for ``clear``
    """

    async def testing():
        tr = TaskResults()

        await tr.add_running_task(task_uid="abc", payload={"some_value": "arbitrary_payload"})
        await tr.add_running_task(task_uid="def", payload={"some_value": "arbitrary_payload"})
        await tr.add_completed_task(task_uid="abc")

        assert len(tr._completed_tasks_time) == 1
        assert len(tr._completed_tasks_data) == 1
        assert len(tr._running_tasks) == 1

        uid = tr.task_results_uid
        await tr.clear()

        assert tr._running_tasks == {}
        assert tr._completed_tasks_time == []
        assert tr._completed_tasks_data == {}

        # UID is not expected to change
        assert tr.task_results_uid == uid

    asyncio.run(testing())


def test_TaskResults_clean_completed_tasks_1():
    """
    TaskResults: tests for ``clean_completed_tasks``
    """

    async def testing():
        tr = TaskResults(retention_time=1)  # Intentionally set short retention time

        await tr.add_completed_task(task_uid="abc")
        assert len(tr._completed_tasks_data) == 1
        assert len(tr._completed_tasks_time) == 1

        await tr.clean_completed_tasks()  # No effect
        assert len(tr._completed_tasks_data) == 1
        assert len(tr._completed_tasks_time) == 1

        ttime.sleep(0.8)

        # 'add_completed_task' is expected to 'clean' tha task list, but there are no expired tasks yet.
        await tr.add_completed_task(task_uid="def", payload={"some_value": 10})
        assert len(tr._completed_tasks_data) == 2
        assert len(tr._completed_tasks_time) == 2

        ttime.sleep(0.5)

        await tr.clean_completed_tasks()  # Should remove the 1st task
        assert len(tr._completed_tasks_data) == 1
        assert len(tr._completed_tasks_time) == 1

        ttime.sleep(0.8)
        await tr.clean_completed_tasks()  # Should remove the 2nd task
        assert len(tr._completed_tasks_data) == 0
        assert len(tr._completed_tasks_time) == 0

    asyncio.run(testing())


def test_TaskResults_clean_completed_tasks_2():
    """
    TaskResults: tests that ``clean_completed_tasks`` is implicitely called when completed task is added.
    """

    async def testing():
        tr = TaskResults(retention_time=1)  # Intentionally set short retention time

        await tr.add_completed_task(task_uid="abc")
        assert len(tr._completed_tasks_data) == 1
        assert len(tr._completed_tasks_time) == 1

        ttime.sleep(1.5)

        # Adds the 2nd task, but removes the 1st (because it is expired)
        await tr.add_completed_task(task_uid="def", payload={"some_value": 10})
        assert len(tr._completed_tasks_data) == 1
        assert len(tr._completed_tasks_time) == 1

        assert tr._completed_tasks_time[0]["task_uid"] == "def"
        assert list(tr._completed_tasks_data.keys())[0] == "def"

    asyncio.run(testing())


def test_TaskResults_get_task_info():
    """
    TaskResults: ``get_task_info``.
    """

    async def testing():
        tr = TaskResults(retention_time=1)  # Intentionally set short retention time

        await tr.add_running_task(task_uid="abc", payload={"some_value": 5})
        await tr.add_running_task(task_uid="def", payload={"some_value": 10})
        await tr.add_completed_task(task_uid="def", payload={"some_value": 20})

        status, payload = await tr.get_task_info(task_uid="abc")
        assert status == "running"
        assert payload == {"some_value": 5}

        status, payload = await tr.get_task_info(task_uid="def")
        assert status == "completed"
        assert payload == {"some_value": 20}

        status, payload = await tr.get_task_info(task_uid="gih")
        assert status == "not_found"
        assert payload == {}

    asyncio.run(testing())
